@@ -15,8 +15,12 @@
 ## Native python imports
 # To store and parse lists from configparser
 import ast
+# To write UTF8 characters in log files
+import codecs
 # To parse config file
 import configparser
+# To use as IDs for each entry in OCR and translation log files
+import datetime
 # For path handling
 import glob
 import os
@@ -26,17 +30,20 @@ import optparse
 import sys
 # For the waiting loop
 import time
-# To display screenshots and select a region
+# To display Tkinter windows in a thread (and be non-blocking so that shortcuts still work)
+import threading
+# To display screenshots and select a region and preprocess the screenshot to enhance letters detection against translucent backgrounds
 if sys.version_info[0] == 2:  # the tkinter library changed it's name from Python 2 to 3.
     import Tkinter
     tkinter = Tkinter #I decided to use a library reference to avoid potential naming conflicts with people's programs.
 else:
     import tkinter
-from PIL import Image, ImageTk
-# To show error message boxes
-from tkinter import messagebox
-# To show textboxes with scrollbars
-import tkinter.scrolledtext
+from PIL import Image, ImageTk, ImageFilter, ImageMath
+# To show error message boxes and textboxes with scrollbars
+from tkinter import messagebox, scrolledtext
+
+## Internal auxiliary scripts
+from _version import __version__
 
 ## External modules
 # For global hotkeys (TODO: find another library to support MacOSX)
@@ -53,16 +60,29 @@ from googletrans import Translator
 
 ### Auxiliary functions
 
+global prev_tk
+prev_tk = None
 class showPILandSelect(object):
     """Display a screenshot in fullscreen and allow to select a rectangle region inside"""
     def __init__(self, pilImage, quitOnSelect=False):
         # Prepare Tkinter fullscreen image/screenshot display
-        root = tkinter.Tk()
+        global prev_tk
+
+        # Create ONE SINGLE root Tk(), else it will fail on subsequent displays
+        if prev_tk is None:
+            prev_tk = tkinter.Tk()
+            prev_tk.overrideredirect(1)
+            prev_tk.withdraw()
+
+        # Create a Toplevel() dialog window, we can create multiple ones
+        root = tkinter.Toplevel()  # use Toplevel() instead of Tk() to reopen the same dialog multiple times, see also: https://github.com/dangillet/PythonFaqFr/blob/master/doc/tkinter.md and https://stackoverflow.com/questions/39458318/how-to-allow-a-tkinter-window-to-be-opened-multiple-times
         self.root = root
-        w, h = root.winfo_screenwidth(), root.winfo_screenheight()
+        # hide root dialog, useless
+        
         root.overrideredirect(1)
+        # Get screen size
+        w, h = root.winfo_screenwidth(), root.winfo_screenheight()
         root.geometry("%dx%d+0+0" % (w, h))
-        root.focus_set()    
         root.bind("<Escape>", lambda e: (e.widget.withdraw(), e.widget.quit()))
         canvas = tkinter.Canvas(root,width=w,height=h)
         canvas.pack()
@@ -86,7 +106,7 @@ class showPILandSelect(object):
         canvas.bind("<B1-Motion>", self.on_move_press)
 
         canvas.create_text(w/2,h/2,fill="red",font="Times 20 italic bold",
-                        text="Please select the region to capture text to translate\n(use mouse left click).\nPress ESCAPE when done.")
+                        text="Please select the region to capture text to translate\n(use mouse left click).%s" % ("\nPress ESCAPE when done." if not quitOnSelect else ""))
 
         # Quit directly after selecting?
         if quitOnSelect:
@@ -94,6 +114,12 @@ class showPILandSelect(object):
         else:
             canvas.bind("<ButtonRelease-1>", self.on_button_release)
 
+        # set window on foreground, on top of others
+        root.focus_set()
+        root.lift()
+        root.attributes('-topmost', 'true')
+
+        # Launch the window
         root.mainloop()
 
     def on_button_press(self, event):
@@ -124,13 +150,16 @@ class showPILandSelect(object):
         else:
             return self.canvas.bbox(self.rect)
 
+    def destroy(self):
+        self.root.destroy()
+
 def showPIL(pilImage):
     # Inspired by code by Neil from https://stackoverflow.com/a/47317411/1121352
     root = tkinter.Tk()
     w, h = root.winfo_screenwidth(), root.winfo_screenheight()
     root.overrideredirect(1)
     root.geometry("%dx%d+0+0" % (w, h))
-    root.focus_set()    
+    root.focus_set()
     root.bind("<Escape>", lambda e: (e.widget.withdraw(), e.widget.quit()))
     canvas = tkinter.Canvas(root,width=w,height=h)
     canvas.pack()
@@ -169,7 +198,7 @@ def selectRegion(sct, config, configFile, quitOnSelect=False):
     # Get the selected rectangle's coordinates
     rectcoords = appregionselect.get_rect_coords()
     # Need to manually destroy Tk root window, else if we try to display it twice it will fail with a weird exception
-    appregionselect.root.destroy()
+    #appregionselect.destroy()
     # Save in config
     if rectcoords is not None:
         if not 'INTERNAL' in config:
@@ -178,7 +207,94 @@ def selectRegion(sct, config, configFile, quitOnSelect=False):
         with open(configFile, 'w') as cfg:
             config.write(cfg)
 
+global prev_root
+prev_root = None
 def translateRegion(sct, config, configFile):
+    class TranslationBox(threading.Thread):
+        def __init__(self, ocrtext, transtext):
+            threading.Thread.__init__(self)
+            self.ocrtext = ocrtext
+            self.transtext = transtext
+            self.config = config
+            self.configFile = configFile
+            self.start()
+        
+        def closeWindow(self):
+            self.root.withdraw()
+            self.root.quit()
+
+        def save_geometry(self, event):
+            """Save size and position of window when moved around to reopen later at the same position"""
+            # Inspired by https://stackoverflow.com/a/43160322/1121352
+            if not 'INTERNAL' in self.config:
+                self.config['INTERNAL'] = {}
+            self.config['INTERNAL']['translationbox_position'] = self.root.geometry()
+            with open(self.configFile, 'w') as cfg:
+                self.config.write(cfg)
+
+        def load_geometry(self):
+            """Load size and position of window from config file"""
+            if 'INTERNAL' in self.config and 'translationbox_position' in self.config['INTERNAL']:
+                self.root.geometry(self.config['INTERNAL']['translationbox_position'])
+
+        def translate(self):
+            # Update ocrtext with the textbox input
+            self.ocrtext = self.txtsrc.get("1.0","end-1c")  # end-1c trick from https://stackoverflow.com/questions/14824163/how-to-get-the-input-from-the-tkinter-text-box-widget
+            # Translate using Google Translate through the googletrans (unofficial) wrapper module
+            self.transtext = gtranslate(self.ocrtext, self.config['DEFAULT']['lang_source_trans'], self.config['DEFAULT']['lang_target'])
+            # Clear up the translation textbox
+            self.txtout.delete("1.0", tkinter.END)
+            # Rewrite the translation textbox content with the new translation
+            self.txtout.insert('1.0', self.transtext)
+
+        def run(self):
+            root = tkinter.Tk()
+            self.root = root
+            # Clean up window on close
+            root.protocol("WM_DELETE_WINDOW", self.closeWindow)
+            # Load position if one is saved in config file
+            self.load_geometry()
+            # Save size and position in config on move
+            root.bind("<Configure>", self.save_geometry)
+            # Title and default window size
+            root.title("pyugt translation")
+            root.geometry('300x500')
+            # Make content resizable + give same weight for both textboxes so they have the same size
+            root.grid_columnconfigure(0, weight=1)
+            root.grid_columnconfigure(1, weight=1)
+            root.grid_columnconfigure(2, weight=1)
+            root.grid_rowconfigure(0, weight=1)
+            root.grid_rowconfigure(1, weight=1)
+            root.grid_rowconfigure(2, weight=1)
+            root.resizable(True, True)
+            # Create the textboxes (with scrollbars)
+            self.txtsrc = scrolledtext.ScrolledText(root, undo=True)
+            self.txtout = scrolledtext.ScrolledText(root, undo=True)
+            # Place the textboxes vertically
+            self.txtsrc.grid(row=0)
+            self.txtout.grid(row=1)
+            # Insert the text (translation and original)
+            self.txtsrc.insert('1.0', self.ocrtext)  # original OCR'ed text in the 1st textbox
+            self.txtout.insert('1.0', self.transtext)  # translated text in the 2nd textbox
+            # Create a button to allow user to manually edit the OCR'ed text and manually force translation again
+            self.buttontrans = tkinter.Button(root, text='Translate again', command=self.translate)
+            self.buttontrans.grid(row=3)
+            # Set window on foreground and stay on top of others
+            root.focus_set()
+            root.lift()
+            root.attributes('-topmost', 'true')
+            # Show window
+            root.mainloop()
+
+    def gtranslate(ocrtext, langsource_trans, langtarget):
+        """Translate using Google Translate through the googletrans (unofficial) wrapper module"""
+        translator = Translator()
+        transobj = translator.translate(ocrtext, src=langsource_trans, dest=langtarget)
+        transtext = transobj.text
+        return transtext
+
+    global prev_root  # allow to check if a previous root window is still open, and close it down
+
     if config['DEFAULT']['debug'] == 'True':
         print('translateRegion triggered')
     # First check a region was set, else raise an error
@@ -197,12 +313,45 @@ def translateRegion(sct, config, configFile):
     # Convert to a PIL Image (else we can't show it on screen)
     img = Image.frombytes("RGB", sct_img.size, sct_img.bgra, "raw", "BGRX")
 
+    # Save screenshot if in debug mode
     if config['DEFAULT']['debug'] == 'True':
         # Get path to temporary image file (need to save to a file to pass onto Tesseract, there's no other way to directly pipe)
         imgtemppath = 'debugtranslate.png'
         # Save to the picture file
         mss.tools.to_png(sct_img.rgb, sct_img.size, output=imgtemppath)
 
+    # Preprocess screenshot to improve OCR accuracy (particularly over translucent backgrounds)
+    if config['DEFAULT']['preprocessing'] == 'True':
+        # Convert to greyscale and upscale image to artificially increase resolution
+        img = img.convert('L').resize((2 * _ for _ in img.size), resample=Image.LANCZOS)
+        if config['DEFAULT']['preprocessing_filters'] != 'None':
+            # Load up the list of filters to apply, from the config file
+            # The filters should be part of PILLOW.ImageFilter
+            imfilters = ast.literal_eval(config['DEFAULT']['preprocessing_filters'])
+            # Apply filters (remove translucent background, increase contrast of the letters, etc)
+            for imfilter in imfilters:
+                img = img.filter(getattr(ImageFilter, imfilter))
+        if config['DEFAULT']['preprocessing_binarize_threshold'] != 'None':
+            # Binarize
+            #fn = lambda x : 255 if x > thresh else 0  # real binarization
+            fn = lambda p : p > thresh and p + 100  # fake binarization, we simply add 100 to the pixels above the threshold, this helps keep some of the variation in the pixels intensities which can be helpful to recognize the letters, from: https://stackoverflow.com/questions/51688973/image-preprocessing-for-ocr-tessaract
+            thresh = int(config['DEFAULT']['preprocessing_binarize_threshold'])
+            try:
+                img = img.point(fn, mode='1')
+            except ValueError as exc:
+                # Exception when the screenshot is too small
+                pass
+        if config['DEFAULT']['preprocessing_invert'] == 'True':
+            # Invert image so that the text is black and background white, this works better with Tesseract according to doc: https://tesseract-ocr.github.io/tessdoc/ImproveQuality
+            img = ImageMath.eval('255 - (a)', a=img.convert('1')).convert('L')  # convert back to greyscale, else can't save a binary image and is buggy for PILLOW and maybe for Tesseract, so better be safe
+
+    # Save preprocessed screenshot if in debug mode
+    if config['DEFAULT']['debug'] == 'True':
+        # Get path to temporary image file (need to save to a file to pass onto Tesseract, there's no other way to directly pipe)
+        imgtemppath = 'debugtranslatepreproc.png'
+        # Save to the picture file
+        img.save(imgtemppath, 'PNG')
+
     # Tesseract OCR to extract text, directly from a PIL image object thanks to the pytesseract wrapper
     ocrtext = pytesseract.image_to_string(img, lang=langsource_ocr, nice=1)
     # TODO: use image_to_boxes or image_to_osd or image_to_data to get position of strings and place them back in place on a screenshot, similarly to what Universal Game Translator does
@@ -215,34 +364,48 @@ def translateRegion(sct, config, configFile):
     #os.system('tesseract -l {imgpath} {srclang} {outputtxt}'.format(imgpath=os.path.abspath(imgtemppath), srclang=langsource, outputtxt='test'))  # alternative way to generate the OCR, by commandline call directly
 
     # Translate using Google Translate through the googletrans (unofficial) wrapper module
-    translator = Translator()
-    transobj = translator.translate(ocrtext, src=langsource_trans, dest=langtarget)
+    if config['DEFAULT']['ocr_only'] == 'True':
+        # Do not translate if ocr_only is enabled
+        transtext = ''
+    else:
+        # Send ocr text to Google Translate to get a translation
+        if config['DEFAULT']['remove_line_returns'] == 'True':
+            # If enabled, remove line returns automatically, so that we consider all sentences to be one (this can help the translator make more sense because it will have more context to work with).
+            ocrtext = ocrtext.replace("\n", "")
+        transtext = gtranslate(ocrtext, langsource_trans, langtarget)
 
     if config['DEFAULT']['debug'] == 'True':
         print('Translated text:')
-        print(transobj.text)
+        print(transtext)
 
-    # Show the result in a window
-    root = tkinter.Tk()
-    root.title("pyugt Translation")
-    root.geometry('300x500')
-    # Make content resizable + give same weight for both textboxes so they have the same size
-    root.grid_columnconfigure(0, weight=1)
-    root.grid_columnconfigure(1, weight=1)
-    root.grid_rowconfigure(0, weight=1)
-    root.grid_rowconfigure(1, weight=1)
-    root.resizable(True, True)
-    # Create the textboxes (with scrollbars)
-    txtsrc = tkinter.scrolledtext.ScrolledText(root)
-    txtout = tkinter.scrolledtext.ScrolledText(root)
-    # Place the textboxes vertically
-    txtsrc.grid(row=0)
-    txtout.grid(row=1)
-    # Insert the text (translation and original)
-    txtsrc.insert('1.0', ocrtext)
-    txtout.insert('1.0', transobj.text)
-    # Show window
-    root.mainloop()
+    # Save OCR'ed text and translation in logs if specified. The same datetime will be written for both, so that they can be matched externally.
+    curdatetime = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
+    if config['DEFAULT']['log_ocr'] != 'None':
+        with codecs.open(config['DEFAULT']['log_ocr'], 'a', 'utf-8-sig') as f:
+            f.write("-> OCR at %s:\n" % curdatetime)
+            f.write(ocrtext)
+            f.write("\n---------------------\n")
+    if config['DEFAULT']['log_translation'] != 'None':
+        with codecs.open(config['DEFAULT']['log_translation'], 'a', 'utf-8-sig') as f:
+            f.write("-> Translation at %s:\n" % curdatetime)
+            f.write(transtext)
+            f.write("\n---------------------\n")
+
+    # Show the translation box
+    if prev_root is not None:
+        # Close previous window if still open
+        # TODO: instead of closing, reinsert text (so we keep window position)
+        try:
+            prev_root.closeWindow()
+        except RuntimeError as exc:
+            # TODO: sometimes when showing a translationbox and then an errorbox, closing the error box will also close the translationbox, hence destroying the root but prev_root will still point to it. If this happens and we try to destroy it again preemptively as we do here, a RuntimeError will be raised, so we just skip. The root cause of this race condition should be fixed instead of this hacky workaround.
+            # To reproduce: do a translation with CTRL+F2, then do it again but on a region with nothing to fail with an error box, and retry until the translation box also closes down when the error box is closed by clicking on the OK button, then finally simply try to translate a text region again and the exception should happen. Maybe it's because we destroy() the errorbox instead of just closing it?
+            if config['DEFAULT']['debug'] == 'True':
+                print('Error when trying to pre-emptively closing prev_root: ')
+                print(exc)
+            pass
+    twindow = TranslationBox(ocrtext, transtext)
+    prev_root = twindow
 
 
 def selectAndTranslateRegion(sct, config, configFile):
@@ -252,9 +415,10 @@ def selectAndTranslateRegion(sct, config, configFile):
 
 def show_errorbox(msg):
     """Show an error box"""
-    root = tkinter.Tk()
+    root = tkinter.Toplevel()
     root.withdraw()
     messagebox.showerror("Error", msg)
+    root.quit()
     root.destroy()  # need to manually destroy the main window, else tk shows an empty window alongside the messagebox... Solution from: https://stackoverflow.com/a/57523108/1121352
 
 def show_errorbox_exception(msg):
@@ -273,7 +437,7 @@ def main():
     configFileArg = options.config
 
     # Print description message on startup
-    print('pyugt - Python Universal Game Translator - started')
+    print('pyugt - Python Universal Game Translator v%s - started' % __version__)
     # Path to current script (to find the config file)
     curpath = os.path.dirname(os.path.abspath(__file__))
     # Load config file
@@ -305,7 +469,7 @@ def main():
     # It's also faster to initialize it only once, per https://python-mss.readthedocs.io/examples.html#benchmark
     sct = mss.mss()
     # Set global hotkeys, loading from config file
-    keyboard.add_hotkey(config['DEFAULT']['hotkey_set_region_capture'], selectRegion, args=(sct, config, configFile))
+    keyboard.add_hotkey(config['DEFAULT']['hotkey_set_region_capture'], selectRegion, args=(sct, config, configFile))  # Do NOT set suppress=True, else this may raise exceptions!
     print('Hit %s to set the region to capture.' % config['DEFAULT']['hotkey_set_region_capture'])
     keyboard.add_hotkey(config['DEFAULT']['hotkey_translate_region_capture'], translateRegion, args=(sct, config, configFile))
     print('Hit %s to translate the region (make sure to close the translation window before requesting another one).' % config['DEFAULT']['hotkey_translate_region_capture'])
